/**
 * TODO: Find a way to properly report errors from recursive operations.
 */
/**
 * The interface for the user-provided objects.
 * TODO: Wait for TS to support `symbol` as index signature type.
 * https://github.com/Microsoft/TypeScript/issues/1863
 * @interface
 */
export interface Data {
    [key: string]: any;
    [index: number]: any;
}
/**
 * User-provided options.
 * @interface
 * @property {string} separator Separator used for the String path format.
 * @property {boolean} enableSpecialParts Determines if registered special parts must processed.
 * @property {object} specialParts An object mapping the corresponding value for a special part.
 * @property {Array<Function>} exclude List of constructors whose instances will be excluded.
 * @property {boolean} force Determines if operations must override existing data.
 * @property {Path.Formats} pathFormat The format for any paths returned.
 * @property {Path.Types} pathType The type of paths to process.
 */
export interface PartialOptions {
    separator?: string;
    enableSpecialParts?: boolean;
    specialParts?: {
        [part: string]: Path.Part;
    };
    exclude?: Function[];
    force?: boolean;
    pathFormat?: Path.Formats;
    pathType?: Path.Types;
}
/**
 * @typedef {string|Array<Path.Part>} Path
 */
export declare type Path = string | Path.Part[];
export declare namespace Path {
    /**
     * Key types (for values computed from RegExp for example).
     * @typedef {string|number} Key
     */
    type Key = string | number;
    /**
     * @typedef {Key|RegExp} Part
     */
    type Part = Key | RegExp;
    /**
     * Available formats for a path.
     * @see Path
     */
    enum Formats {
        String = 0,
        Array = 1
    }
    /**
     * Available types of paths. Leaf paths are paths with no children.
     */
    enum Types {
        Any = 0,
        Leaf = 1
    }
    /**
     * Transforms a path to the desired format.
     * @param {Path} path The path to format.
     * @param {PartialOptions} opts
     * @returns {Path} The formatted path.
     * @throws {TypeError} A part is not compatible with the selected format.
     * @see Options
     */
    function format(path: Path, opts?: PartialOptions): Path;
    /**
     * Returns the keys corresponding to part for an object. Useful to resolve
     * RegExp parts.
     * @param {Data} object
     * @param {Part} part
     * @param {PartialOptions} opts
     * @returns {Array<Key>} The list of keys.
     * @see Options
     */
    function partMatchingKeys(object: Data, part: Part, opts?: PartialOptions): Key[];
    /**
     * Determines if one path is a parent of the other. RegExp parts are not
     * supported.
     *
     * @param {Path} path1 The first path
     * @param {Path} path2 The second path
     * @param {PartialOptions} opts
     * @returns {boolean} The result of the test
     * @see Options
     */
    function arePathsRelated(path1: Path, path2: Path, opts?: PartialOptions): boolean;
    function isPathContainingRegExp(path: Path, opts?: PartialOptions): boolean;
}
/**
 * Determines if a path exists in an object.
 * @param {Data} object
 * @param {Path} path
 * @param {PartialOptions} opts
 * @returns {boolean}
 * @see Options
 */
export declare function exists(object: Data, path: Path, opts?: PartialOptions): boolean;
/**
 * Returns an array of the values corresponding to the path in an object. If no
 * values are found returns a default value if defined (in an array).
 * @param {Data} object
 * @param {Path} path
 * @param {PartialOptions} opts
 * @param {*} defaultVal If defined this value is returned instead of an error.
 * @returns {Array} The list of the values.
 * @throws {UndefinedPropertyError} No values were found.
 * @see Options
 */
export declare function get(object: Data, path: Path, opts?: PartialOptions, defaultVal?: any): any[];
/**
 * Sets a value in an object. Missing properties are created if `force` option
 * is enabled excepted for excluded objects. If no values can be set during the
 * operation, an error is thrown.
 * @param {Data} object
 * @param {Path} path
 * @param {*} value The value to place.
 * @param {PartialOptions} opts
 * @throws {TypeError} The given part cannot be used to create an undefined property.
 * @throws {UndefinedPropertyError} The given object cannot have additional properties.
 * @see Options
 */
export declare function put(object: Data, path: Path, value: any, opts?: PartialOptions): void;
/**
 * Deletes the matching keys in an object.
 * @param {Data} object
 * @param {Path} path
 * @param {PartialOptions} opts
 * @throws {UndefinedPropertyError} The path doesn't lead to any deletable properties.
 * @see Options
 */
export declare function remove(object: Data, path: Path, opts?: PartialOptions): void;
/**
 * Returns the list of available paths for an abject.
 * @param {Data} object
 * @param {PartialOptions} opts
 * @returns {Array<Path>}
 * @see Options
 */
export declare function paths(object: Data, opts?: PartialOptions): Path[];
export default class DotT {
    /**
     * Options passed to each call.
     * @implements {Options}
     * @readonly
     * @private
     */
    private readonly options;
    /**
     * The object the instance works on.
     * @implements {Data}
     * @readonly
     * @private
     */
    private readonly object;
    /**
     * Creates a DotT object. DotT objects are simple wrappers for the main DotT
     * functions.
     * @param {Data} object
     * @param {PartialOptions} opts
     * @see Options
     */
    constructor(object?: Data, opts?: PartialOptions);
    /**
     * Access the object.
     */
    readonly value: Data;
    /**
     * @see exists
     */
    exists(path: Path): boolean;
    /**
     * @see get
     */
    get(path: Path, defaultVal?: any): any;
    /**
     * @returns {DotT} Self
     * @see put
     */
    put(path: Path, value: any): DotT;
    /**
     * @returns {DotT} Self
     * @see remove
     */
    remove(path: Path): DotT;
    /**
     * @see paths
     */
    paths(): Path[];
}
